#include "mpi.h"

/* for fprintf */
#include "stdio.h"

/* for malloc */
#include "stdlib.h"

/* for getpid() */
#if defined(HAVE_UNISTD_H)
#  include "sys/types.h"
#  include "unistd.h"
#endif

#if defined(HAVE_STRING_H)
#  include <string.h>
#endif

#if defined(HAVE_LONG_LONG_INT)
#  define LARGEST_TYPE long long int
#  define LARGEST_MPI_TYPE MPI_LONG_LONG_INT
#  define FMT_SPEC "%lld"
#else /* fall back to 'long int' */
#  define LARGEST_TYPE long int
#  define LARGEST_MPI_TYPE MPI_LONG
#  define FMT_SPEC "%ld"
#endif

static void err_and_die(char *msg)
{
  fprintf(stderr, "%s", msg);fflush(stderr);
  MPI_Abort(MPI_COMM_WORLD, 1);
}

/* This  simple test checks that MPI_Abort kills all processes 
 * There are two interesting cases:
 * masternode == 0
 * masternode != 0
 */
int main( int argc, char **argv )
{
  int node, size, i, ret;
  int masternode = 0;
  FILE *pidfile = NULL;
  LARGEST_TYPE my_pid;
  LARGEST_TYPE *buf = NULL;

  MPI_Init(&argc, &argv);

  MPI_Comm_rank(MPI_COMM_WORLD, &node);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  /* Check for -altmaster */
  for (i=1; i<argc; i++) {
    if (argv[i] && strcmp( "-altmaster", argv[i] ) == 0) {
      masternode = size-1;
    }
    else if (node == masternode && argv[i] && strcmp( "-pidfile", argv[i] ) == 0) {
      if (i+1 >= argc || NULL == argv[i+1]) {
        err_and_die("error, must give a value for -pidfile argument\n");
      }

      pidfile = fopen(argv[i+1], "w");
      if (!pidfile) err_and_die("unable to open given pidfile!\n");
      ++i; /* advance past pidfile value */
    }
  }


  if (node == masternode) {
    if (!pidfile) err_and_die("must specify a pidfile\n");

    buf = malloc(sizeof(my_pid)*size);
    if (!buf) err_and_die("unable to allocate pid buffer\n");
  }

  my_pid = getpid();
  ret = MPI_Gather(&my_pid, 1, LARGEST_MPI_TYPE,
                   buf, 1, LARGEST_MPI_TYPE,
                   masternode, MPI_COMM_WORLD);
  if (ret) err_and_die("MPI_Gather failed\n");

  if (node == masternode) {
    /* write out the pid list for the test harness to use */
    for (i = 0; i < size; ++i) {
      fprintf(pidfile, FMT_SPEC "\n", buf[i]);
    }
    fclose(pidfile);

    MPI_Abort(MPI_COMM_WORLD, 99);
  }
  else {
    /* barrier will hang since masternode never calls */
    MPI_Barrier(MPI_COMM_WORLD);
  }

  MPI_Finalize();
  return 0;
}
